import sys
from pathlib import Path

# aggiunge ../src al PYTHONPATH
ROOT = Path(__file__).resolve().parents[1]
SRC = ROOT / "src"
sys.path.insert(0, str(SRC))


import unittest
from pucktrick.duplicated import *
from pucktrick.utils import * 
import pandas as pd
import numpy as np
from pandas.testing import assert_frame_equal


class Test_outlies(unittest.TestCase):
 
 def test_duplicateClassExtended(self):
     percentage=0.5
     num_rows = 20
     selection_criteria = "f4=='apple' or f3=='apple'"
     selection_criteria='all'
     fake_df=create_fake_table(num_rows)
     column=["f3"]
     if selection_criteria=='all':
          old_len=len(fake_df)
     else:
         old_len=len(fake_df.query(selection_criteria))
     print("original")
     print(fake_df)
     strategy =  {'affected_features': column, 'selection_criteria': selection_criteria, 'function':None, 'percentage': percentage,'mode': 'new', 'perturbate_data': {'distribution': 'random', 'param': None, 'value': None, 'condition_logic': None}}
     error,noise_df=duplicate(fake_df,strategy)
     print("prima modifica")
     print(noise_df)
     percentage=0.7
     strategy =  {'affected_features': column, 'selection_criteria': selection_criteria, 'function':None, 'percentage': percentage,'mode': 'extended', 'perturbate_data': {'distribution': 'random', 'param': None, 'value': None, 'condition_logic': None}}
     error,noise1_df=duplicate(noise_df,strategy,fake_df)
     if error==0:
            noise_df=noise1_df
     print("seconda modifica")
     print(noise_df)
     
     if selection_criteria == "all" or selection_criteria is None:
          new_len=len(noise_df)
          print(new_len)
          diff=round((new_len-old_len)/old_len,1)
          print(diff)
     else:
          #selection_criteria,isAffected=affectedSelected(column,selection_criteria)
          #if isAffected==1:
          #      df_pref2=fake_df[column].add_prefix("puck_")
          #      fake_df=pd.concat([fake_df,df_pref2],axis=1)
          #      noise_df=pd.concat([noise_df,df_pref2],axis=1)
          new_len=len(noise_df.query(selection_criteria)[column])
          diff=round((new_len-old_len)/old_len,1)
          print(diff)
        
        # Verifica che il numero di modifiche sia corretto
     self.assertLessEqual(abs(diff-percentage),0.1)

if __name__ == "__main__":
    unittest.main()



from pucktrick.utils import *

def duplicate(train_df, strategy=None, original_df=None ):

    affected_features = strategy.get("affected_features")
    selection_criteria = strategy.get("selection_criteria")
    percentage = strategy.get("percentage")
    mode = strategy.get("mode")
    function=strategy.get('function')
    perturbate_config = strategy.get("perturbate_data", {})
    distr1 = perturbate_config.get("distribution")
    value = perturbate_config.get("value", [])
    condition_logic = perturbate_config.get("condition_logic", None)
    params = perturbate_config.get("param", {})
    number=None
    isAffectedSelected=0
    if affected_features and not isinstance(affected_features, list):
        affected_features = [affected_features]
    
    if selection_criteria=='all':
            if original_df is None:
                original_df = train_df.copy()
                original1_df = train_df.copy()

            else:
                original1_df=original_df.copy()
    else:
        if mode=="extended":
            selection_criteria,isAffectedSelected=affectedSelected(affected_features,selection_criteria)
        else:
           isAffectedSelected=0 
        if isAffectedSelected==1:
            if original_df is not None:
                df_pref=train_df[affected_features].add_prefix("puck_")
                df_pref2=original_df[affected_features].add_prefix("puck_")
                original_df=pd.concat([original_df,df_pref2],axis=1)
                train_df=pd.concat([train_df,df_pref],axis=1)
                
        if original_df is None:
                original_df = train_df.copy()
                original1_df = train_df.copy()
        else:
            original1_df=train_df.copy()
            original_df=original_df.query(selection_criteria).copy() 
        train_df=train_df.query(selection_criteria).copy()
    if function in [
        "class",
        "upper_lower",
        "replace_punctuation",
        "remove_replace",
        "abbreviate_text",
        "shuffle_words",
    ]:
        # Costruzione dei filtri
        filters = []

        if value is None or not any(value):
            filters = [
                (original_df[col].notnull() if mode == "extended" else train_df[col].notnull())
                for col in affected_features
            ]
        else:
            for col, val in zip(affected_features, value):
                if val is not None:
                    filters.append(
                        (original_df[col] == val) if mode == "extended" else (train_df[col] == val)
                    )
                else:
                    filters.append(
                        (original_df[col].notnull()) if mode == "extended" else (train_df[col].notnull())
                    )

        # Combina i filtri con logica "or" o "and"
        if condition_logic == "and":
            combined_filter = pd.Series(True, index=train_df.index if mode != "extended" else original_df.index)
            for f in filters:
                combined_filter &= f
        elif condition_logic == "or":
            combined_filter = pd.Series(False, index=train_df.index if mode != "extended" else original_df.index)
            for f in filters:
                combined_filter |= f
        elif condition_logic is None:
            # Default: "and" logic
            combined_filter = pd.Series(True, index=train_df.index if mode != "extended" else original_df.index)
            for f in filters:
                combined_filter &= f
        else:
            raise ValueError(f"Unsupported condition_logic '{condition_logic}'")

        # Filtrare le righe
        target_df = original_df if mode == "extended" else train_df
        filtered_df = target_df[combined_filter].reset_index(drop=True)
        rowsToChange = int(len(filtered_df) * percentage)
        if mode == "extended":
            old_duplicated = len(train_df) - len(original_df)
            new_rowsToChange = rowsToChange - old_duplicated
      
        if new_rowsToChange <= 0:
                return 1,train_df
        percentage = new_rowsToChange / len(filtered_df)
        sampled_indices = sample_indices(distr1, filtered_df, percentage, params,number=None)
        sampled_rows = filtered_df.iloc[sampled_indices].copy()
        for col in affected_features:
            if selection_criteria == "shuffle_words":
                sampled_rows[col] = sampled_rows[col].astype(str).apply(shuffle_words)
            elif selection_criteria == "abbreviate_text":
                sampled_rows[col] = sampled_rows[col].astype(str).apply(abbreviate_text)
            elif selection_criteria == "replace_punctuation":
                sampled_rows[col] = sampled_rows[col].astype(str).apply(replace_punctuation)
            elif selection_criteria == "remove_replace":
                sampled_rows[col] = sampled_rows[col].astype(str).apply(remove_or_replace)
            elif selection_criteria == "upper_lower":
                sampled_rows[col] = random_upper_lower(sampled_rows[col])

        noise_df = pd.concat([train_df, sampled_rows], ignore_index=True)

    if mode == "extended":
            rowsToChange = len(original_df) * percentage
            
            if selection_criteria=='all':
                 old_duplicated = original_df.duplicated().sum()
            else:
                 old_duplicated = len(original1_df.query(selection_criteria))-len(original_df)
            #new_duplicated = train_df.duplicated().sum()
            #new_duplicated =len(train_df)
            #diff = new_duplicated - old_duplicated
            #new_rowsToChange = rowsToChange - diff
            new_rowsToChange = rowsToChange - old_duplicated
           
            if new_rowsToChange <= 0:
                return 1,train_df
            new_percentage = new_rowsToChange / len(original_df)
            noise_df = train_df.copy()
            sampled_indices = sample_indices(distr1, original_df, new_percentage, number, params)
            sampled_rows = original_df.iloc[sampled_indices]
            #noise_df = pd.concat([noise_df, sampled_rows], ignore_index=True)
            noise_df=pd.concat([original1_df,sampled_rows],ignore_index=True)
            
    else:
            noise_df = original1_df.copy()
            sampled_indices = sample_indices(distr1, train_df, percentage, number, params)
            sampled_rows = train_df.iloc[sampled_indices]
            noise_df = pd.concat([noise_df, sampled_rows], ignore_index=True)
    if "date" in noise_df.columns:
        noise_df["date"] = noise_df["date"].astype(str).str.replace(r"\s00:00:00$", "", regex=True)
    
    if isAffectedSelected==1:
        puck_cols=[c for c in noise_df.columns if c.startswith("puck_")]
        noise_df=noise_df.drop(columns=puck_cols)
        original1_df=original1_df.drop(columns=puck_cols)
   
    return 0,noise_df




from pucktrick.utils import *
import re

#modificare quando il selection criteria è uguale alla colonna

def is_int_categorical(series: pd.Series,
                       max_unique: int = 10,
                       max_domain_size: int = 20,
                       max_unique_ratio: float = 0.05,
                       require_nonnegative: bool = True) -> bool:
    """
    Heuristic: decide if an integer-typed column is actually categorical (coded as ints).
    """
    if not pd.api.types.is_integer_dtype(series):
        return False
    
    s = series.dropna()
    if s.empty:
        return False
    
    nunq = s.nunique()
    if nunq <= max_unique:
        return True
    
    if nunq <= max_domain_size and nunq / len(s) <= max_unique_ratio:
        return True
    
    vals = np.sort(s.unique())
    if require_nonnegative and vals[0] < 0:
        return False
    
    # consecutive small domain?
    if len(vals) <= max_domain_size and np.all(np.diff(vals) == 1):
        return True
    
    return False

def outlier(train_df, strategy, original_df=None):
    affected_features = strategy.get("affected_features")
    selection_criteria = strategy.get("selection_criteria")
    percentage = strategy.get("percentage")
    mode = strategy.get("mode")
    perturbate_config = strategy.get("perturbate_data", {})
    distr1 = perturbate_config.get("distribution")
    values = perturbate_config.get("value", [])
    condition_logic = perturbate_config.get("condition_logic", None)
    params=perturbate_config.get("param", {})
    isAffectedSelected=0
    if selection_criteria == 'all':
        noisy_df = train_df.copy()
        if original_df is None:
            original_df = train_df.copy()
            original1_df= train_df.copy()
        else:
            original1_df=original_df.copy()
    else:
        if mode=="extended":
            selection_criteria,isAffectedSelected=affectedSelected(affected_features,selection_criteria)
        else:
           isAffectedSelected=0 
        if isAffectedSelected==1:
            if original_df is not None:
                    df_pref2=original_df[affected_features].add_prefix("puck_")
                    original_df=pd.concat([original_df,df_pref2],axis=1)
                    train_df=pd.concat([train_df,df_pref2],axis=1)
            
        if original_df is None:
            original1_df=  train_df.copy()
            original_df = train_df.query(selection_criteria).copy()         
        else:
            original1_df=original_df.copy()
            original_df=original_df.query(selection_criteria).copy()
       
        train_df=train_df.query(selection_criteria).copy()
        noisy_df=train_df.copy()
        
    if affected_features and not isinstance(affected_features, list):
        affected_features = [affected_features]
    if not values:
        filtered_df = noisy_df
    else:
        row_filters = []
        for feature, value in zip(affected_features, values):
            if value is None:
                row_filters.append(train_df[feature].notnull())
            else:
                row_filters.append(train_df[feature] == value)

        if condition_logic == "and":
            combined_filter = row_filters[0]
            for f in row_filters[1:]:
                combined_filter &= f
        else:
            combined_filter = row_filters[0]
            for f in row_filters[1:]:
                combined_filter |= f

        filtered_df = noisy_df[combined_filter]
    total_rows_to_change = int(len(filtered_df) * percentage)
#    print("total_rows_to_change", str(total_rows_to_change))

    if mode == "extended" and original_df is not None:
        already_modified_mask = train_df[affected_features].ne(original_df[affected_features]).any(axis=1)
        already_modified_count = already_modified_mask.sum()
        rows_to_change = total_rows_to_change - already_modified_count 
        modifiable_indices = filtered_df[~already_modified_mask.reindex(filtered_df.index, fill_value=False)]
        if rows_to_change <= 0 or len(modifiable_indices) == 0:
            return 1,train_df
        number=rows_to_change
      #  print("numero righe da modificare extended", number )
        percentage=None
        sampled_indices = sample_indices(distr1, modifiable_indices,percentage, number, params)
        sampled_indices = modifiable_indices.iloc[sampled_indices].index
    else:
        number=total_rows_to_change
       # print("numero righe da modificare new", str(number) )
        percentage=None
        sampled_indices = sample_indices(distr1, filtered_df,percentage, number, params)
        sampled_indices = filtered_df.iloc[sampled_indices].index
    
    for feature in affected_features:
        if pd.api.types.is_string_dtype(noisy_df[feature]):
            for i in sampled_indices:
                noisy_df.loc[i, feature] = "puck was here"

        elif pd.api.types.is_integer_dtype(noisy_df[feature]):
            if is_int_categorical(noisy_df[feature]):
                current_max = noisy_df[feature].max(skipna=True)
                current_max=current_max+1
                for i in sampled_indices:
                    noisy_df.loc[i, feature] = current_max
            else:
                mean = np.mean(noisy_df[feature])
                std_dev = np.std(noisy_df[feature])
                new_upper = round(mean + 4 * std_dev, 0)
                new_lower = round(mean - 4 * std_dev, 0)

                for i in sampled_indices:
                    if np.random.rand() > 0.5:
                        noisy_df.loc[i, feature] = random.randint(int(mean + 3 * std_dev), int(new_upper))
                    else:
                        noisy_df.loc[i, feature] = random.randint(int(new_lower), int(mean - 3 * std_dev))

        elif pd.api.types.is_datetime64_any_dtype(noisy_df[feature]):
            valid_dates = train_df[feature].dropna()
            if valid_dates.empty:
                continue
            min_date = valid_dates.min()
            max_date = valid_dates.max()

            for i in sampled_indices:
                random_days = random.randint(0, (max_date - min_date).days)
                noisy_df.loc[i, feature] = min_date + timedelta(days=random_days)

        elif pd.api.types.is_float_dtype(noisy_df[feature]):
            mean = np.mean(noisy_df[feature])
            std_dev = np.std(noisy_df[feature])
            new_upper = mean + 4 * std_dev
            new_lower = mean - 4 * std_dev

            for i in sampled_indices:
                if np.random.rand() > 0.5:
                    noisy_df.loc[i, feature] = np.random.uniform(mean + 3 * std_dev, new_upper)
                else:
                    noisy_df.loc[i, feature] = np.random.uniform(new_lower, mean - 3 * std_dev)

        else:
            raise ValueError(f"Unsupported column type for feature '{feature}'.")
    if isAffectedSelected==1:
        puck_cols=[c for c in noisy_df.columns if c.startswith("puck_")]
        noisy_df=noisy_df.drop(columns=puck_cols)
        original1_df=original1_df.drop(columns=puck_cols)
        
    original1_df.loc[noisy_df.index, noisy_df.columns]=noisy_df
    return 0,original1_df


    
    




from pucktrick.utils import *
#### 
#
# da controllare quando il filtro è sui valori da cancellare
####
def missing(train_df,  strategy=None, original_df=None):
    
    affected_features = strategy.get("affected_features")
    selection_criteria = strategy.get("selection_criteria")
    percentage = strategy.get("percentage")
    mode = strategy.get("mode")
    perturbate_config = strategy.get("perturbate_data", {})
    distr1 = perturbate_config.get("distribution")
    value = perturbate_config.get("value", [])
    condition_logic = perturbate_config.get("condition_logic", None)
    params= perturbate_config.get("param", {})
    isAffectedSelected=0
    if affected_features and not isinstance(affected_features, list):
        affected_features = [affected_features]

    if value is None:
        value = [None] * len(affected_features)
        
    if selection_criteria=='all':
        noise_df = train_df.copy()
        if original_df is None:
            original_df = train_df.copy()
            original1_df= train_df.copy()
        else:
            original1_df=original_df.copy()

    else:
        if mode=="extended":
            selection_criteria,isAffectedSelected=affectedSelected(affected_features,selection_criteria)
        else:
           isAffectedSelected=0 
        if isAffectedSelected==1:
            if original_df is not None:
                df_pref2=original_df[affected_features].add_prefix("puck_")
                original_df=pd.concat([original_df,df_pref2],axis=1)
                train_df=pd.concat([train_df,df_pref2],axis=1)
         
        if original_df is None:
            original1_df=  train_df.copy()
            original_df = train_df.query(selection_criteria).copy()         
        else:
            original1_df=original_df.copy()
            original_df=original_df.query(selection_criteria).copy()
        train_df=train_df.query(selection_criteria).copy()
        noise_df=train_df.copy()
            

    if mode == 'extended':
        filters_original = []
        filters_train = []

        for col, val in zip(affected_features, value):
            if val is not None:
                filters_original.append(original_df[col] == val)
                filters_train.append(train_df[col] == val)
            else:
             
                filters_original.append(None)
                filters_train.append(None)

  
        combined_filter_original = None
        combined_filter_train = None

        for filter_original in filters_original:
            if filter_original is not None:
                if combined_filter_original is None:
                    combined_filter_original = filter_original
                else:
                    if condition_logic == "and":
                        combined_filter_original &= filter_original
                    elif condition_logic == "or":
                        combined_filter_original |= filter_original

        for filter_train in filters_train:
            if filter_train is not None:
                if combined_filter_train is None:
                    combined_filter_train = filter_train
                else:
                    if condition_logic == "and":
                        combined_filter_train &= filter_train
                    elif condition_logic == "or":
                        combined_filter_train |= filter_train

        if combined_filter_original is not None:
            origin_column_df = original_df[combined_filter_original]
        else:
            origin_column_df = original_df

        if combined_filter_train is not None:
            target_df = train_df[combined_filter_train]
        else:
            target_df = train_df

        total_rows_to_change = round(len(origin_column_df) * percentage)
        already_null_counts = train_df[affected_features].isnull().all(axis=1).sum()
        new_rows_to_change = total_rows_to_change - already_null_counts

        if new_rows_to_change > 0:
            modifiable_df = target_df[~target_df[affected_features].isnull().all(axis=1)]
            number= new_rows_to_change
            percentage=None

            sampled_indices = sample_indices(distr1, modifiable_df,percentage, number, params)
            sampled_rows = modifiable_df.iloc[sampled_indices].index
            for col in affected_features:
                noise_df.loc[sampled_rows, col] = np.nan
     

    else:
        filters = []

        for col, val in zip(affected_features, value):
            if val is not None:
                filters.append(train_df[col] == val)
            else:
   
                filters.append(None)
        #print("Filtri:", filters)
        if selection_criteria == "all":
            combined_filter = pd.Series(True, index=train_df.index)
        else:
            combined_filter = None
            for filter_condition in filters:
                if filter_condition is not None:
                    if combined_filter is None:
                       combined_filter = filter_condition
                    else:
                        if condition_logic == "and":
                            combined_filter &= filter_condition
                        elif condition_logic == "or":
                            combined_filter |= filter_condition
            if combined_filter is None:
              combined_filter = pd.Series(True, index=train_df.index)
       
    
        if combined_filter is not None:
            column_df = train_df[combined_filter]
        else:
            column_df = train_df

        total_rows_to_change = int(len(column_df) * percentage)
        if total_rows_to_change <= 0:
            return 1,train_df
        number= total_rows_to_change
        percentage=None
        sampled_indices = sample_indices(distr1, column_df, percentage,number, params)
        sampled_rows = column_df.iloc[sampled_indices].index
        #print(sampled_rows)
        for col in affected_features:
            noise_df.loc[sampled_rows, col] = np.nan
    if isAffectedSelected==1:
        puck_cols=[c for c in noise_df.columns if c.startswith("puck_")]
        noise_df=noise_df.drop(columns=puck_cols)
        original1_df=original1_df.drop(columns=puck_cols)
    
    original1_df.loc[noise_df.index, noise_df.columns]=noise_df
    return 0,original1_df



import unittest
from pucktrick.utils import *
import pandas as pd
import numpy as np
from pandas.testing import assert_frame_equal
import random
import json
from datetime import timedelta
from dateutil import parser
import string
import re

def create_fake_table(num_rows=1000):
  # Generate data for each column
  f1 = np.random.uniform(-100, 100, num_rows)  # Continuous values between -100 and 100
  f2 = np.random.randint(-100, 101, num_rows)  # Discrete values between -100 and 100
  f3 = np.random.choice(['apple', 'banana', 'cherry'], num_rows)  # String values
  f4 = np.random.choice(['apple', 'banana', 'cherry'], num_rows)  # String values
  f5 = np.random.choice([0, 1,2,3,4,5], num_rows)  # Binary values (0 or 1)
  delta_days = (pd.to_datetime('2030-12-31') - pd.to_datetime('1970-01-01')).days
  random_days = np.random.randint(0, delta_days, num_rows)
  random_dates = pd.to_datetime('1970-01-01') + pd.to_timedelta(random_days, unit='D')
  date = random_dates  # Restituisce direttamente una Series di tipo datetime.date
  # date between 1970-01-01 and 2030-12-31
  target = np.random.choice([0, 1], num_rows)  # Binary target (0 or 1)

  # Create the DataFrame
  df = pd.DataFrame({
    'f1': f1,
    'f2': f2,
    'f3': f3,
    'f4': f4,
    'f5': f5,
    'date' : date,
    'target': target
  })
  df['date'] = pd.to_datetime(df['date'])
  return df

def generateSubdf(original_df, train_df,column,percentage ):
    rowsToChange=len(original_df[column])*percentage
    dif = original_df[column] != train_df[column]
    diff_number = dif.sum()
    new_rowsToChange=rowsToChange-diff_number
    if new_rowsToChange<=0:
      newPercentage=0
      return train_df,newPercentage
    noise_df= train_df.copy()
    noise_df['id1'] = noise_df.index
    or_df=original_df.copy()
    or_df['id1'] = noise_df.index
    mask = or_df[column] == noise_df[column]
    new_df = noise_df[mask]
    new_df = new_df.reset_index(drop=True)
    newPercentage=new_rowsToChange/len(new_df)
    return new_df,newPercentage

def mergeDataframe(noise_df,modified_df):
    modified_df=modified_df.set_index('id1')
    noise_df.update(modified_df)
    noise_df = noise_df.drop('id1', axis=1)
    return noise_df

def generate_random_value(lower, upper):
        return np.random.uniform(lower, upper)

def generate_random_value_discrete(lower, upper):
        return np.random.randint(lower, upper)

def sampleList(percentage, data_length):
    num_samples = int(percentage * data_length) 
    return np.random.choice(data_length, size=num_samples, replace=False)
  
def sample_indices(distribution, data_length, percentage=None, number=None, params=None):
    if percentage is None:
      num_samples = number
    elif number is None:
      num_samples = int(len(data_length) * percentage)  # use data_length directly

    params = params or {}
    #print(percentage)

    if distribution == "random":
        # Change: use len(data_length) to get the size for random sampling
        return np.random.choice(len(data_length), num_samples, replace=False)

    elif distribution == "uniform":
        min_val = params.get("min", 0)
        max_val = params.get("max", len(data_length) - 1)  # data_length
        return np.random.uniform(min_val, max_val, num_samples).astype(int)

    elif distribution == "normal":
        mu = params.get("mu", len(data_length) / 2)  # data_length
        sigma = params.get("sigma", len(data_length) / 6)  # data_length
        samples = np.random.normal(mu, sigma, num_samples)
        samples = np.clip(samples, 0, len(data_length) - 1)  # data_length
        return samples.astype(int)

    elif distribution == "exponential":
        lambd = params.get("lambda", 1.0)
        samples = np.random.exponential(1 / lambd, num_samples)
        samples = (samples / np.max(samples)) * (len(data_length) - 1)  # data_length
        return samples.astype(int)

    else:
        raise ValueError(f"Distribuzione '{distribution}' non supportata.")

def noiseBinaryNew(train_df, target, distribution, percentage , number, params):
    noise_df = train_df
    indices_to_modify= sample_indices(distribution, train_df, percentage, number, params)
    for i in indices_to_modify:
        noise_df.iloc[i,noise_df.columns.get_loc(target)] = 1 - noise_df.iloc[i,noise_df.columns.get_loc(target)]  
    return noise_df

def noiseBinaryExt(original_df, train_df, target, distribution,percentage,number,params):
    noise_df = train_df.copy()
    noise_df['id1'] = noise_df.index  
    
    new_df, newPercentage = generateSubdf(original_df, noise_df, target, percentage)
   
    if newPercentage == 0:
        return train_df
    modified_df = noiseBinaryNew(new_df, target, distribution, newPercentage,number, params)
    noise_df= mergeDataframe(noise_df, modified_df)
    return noise_df

def noiseCategoricalStringNewExistingValues(train_df, column, distribution, percentage , number, params):
    #extracted_indices = sampleList(percentage, len(train_df[column]))
    extracted_indices = sample_indices(distribution, train_df[column], percentage=None, number=None, params=None)
    noise_df = train_df.copy()
    unique_values = noise_df[column].unique()

    for i in extracted_indices:
        while True:
            new_value = np.random.choice(unique_values)
            if new_value != noise_df.loc[i, column]:
                noise_df.loc[i, column] = new_value
                break
    return noise_df

def noiseCategoricalIntNewExistingValues(train_df, target,  distribution, percentage , number, params):
    noise_df = train_df.copy()
    #indices_to_modify = sampleList(percentage, len(train_df[target]))
    indices_to_modify= sample_indices(distribution, train_df[target], percentage=None, number=None, params=None)
    unique_values = train_df[target].unique()
    for i in indices_to_modify:
        current_value = noise_df.at[i, target]
        new_value = np.random.choice([v for v in unique_values if v != current_value])
        noise_df.at[i, target] = new_value
    return noise_df

def noiseCategoricalIntExtendedExistingValues(original_df, train_df, target, distribution,percentage,number,params):
    noise_df = train_df.copy()
    noise_df['id1'] = range(len(noise_df))  # Aggiunta della colonna 'id1'
    new_df, newPercentage = generateSubdf(original_df, noise_df, target, percentage)
    if newPercentage == 0:
        return train_df
    modified_df = noiseCategoricalIntNewExistingValues(new_df, target, distribution, newPercentage,number, params)
    noise_df = mergeDataframe(noise_df, modified_df)
    return noise_df

def noiseCategoricalStringNewFakeValues(train_df, column, distribution,percentage,number,params):
    #extracted_list = sampleList(percentage, len(train_df))
    extracted_list= sample_indices(distribution, train_df, percentage=None, number=None, params=None) # Cambia qui per selezionare casualmente
    noise_df = train_df.copy()
    for i in extracted_list:  # Usa extracted_list per modificare solo i campioni
        noise_df.loc[i, column] = ''.join(np.random.choice(list('abcdefghijklmnopqrstuvwxyz'), size=5))
    return noise_df

def noiseCategoricalStringExtendedFakeValues(original_df, train_df, column, distribution,percentage,number,params):
    noise_df = train_df.copy()
    noise_df['id1'] = range(len(noise_df))
    new_df, newPercentage = generateSubdf(original_df, noise_df, column, percentage)
    if newPercentage == 0:
        return train_df

    modified_df = noiseCategoricalStringNewFakeValues(new_df, column, distribution,newPercentage,number,params)
    noise_df = mergeDataframe(noise_df, modified_df)

    # Rimuovi la colonna 'id1' temporanea se esiste
    if 'id1' in noise_df.columns:
        noise_df = noise_df.drop('id1', axis=1)

    return noise_df

# Funzione per aggiungere rumore discreto (normal)
def noiseDiscreteNew(train_df, column, distribution,percentage,number,params):
    #extracted_indices = sampleList(percentage, len(train_df[column]))
    extracted_indices= sample_indices(distribution, train_df[column], percentage=None, number=None, params=None)
    noise_df = train_df.copy()
    min_value = noise_df[column].min()
    max_value = noise_df[column].max()

    for i in extracted_indices:
        while True:
            new_value = random.randint(min_value, max_value)
            if new_value != noise_df.loc[i, column]:
                noise_df.loc[i, column] = new_value
                break
    return noise_df

# Funzione per aggiungere rumore continuo (normal)
def noiseContinueNew(train_df, column, distribution,percentage,number,params):
    #extracted_indices = sampleList(percentage, len(train_df[column]))
    extracted_indices= sample_indices(distribution, train_df[column], percentage=None, number=None, params=None)
    noise_df = train_df.copy()
    min_value = noise_df[column].min()
    max_value = noise_df[column].max()

    for i in extracted_indices:
        while True:
            new_value = random.uniform(min_value, max_value)
            if new_value != noise_df.loc[i, column]:
                noise_df.loc[i, column] = new_value
                break
    return noise_df

def outlierContinuosNew3Sigma(train_df, column, distribution,percentage,number,params):
    #extracted_list = sampleList(percentage, len(train_df[column]))
    extracted_list= sample_indices(distribution, train_df[column], percentage=None, number=None, params=None)
    noise_df = train_df.copy()
    mean = np.mean(noise_df[column])
    std_dev = np.std(noise_df[column])
    upper_bound = mean + 3 * std_dev
    lower_bound = mean - 3 * std_dev
    new_lower_limit = mean - 4 * std_dev
    new_upper_limit = mean + 4 * std_dev

    for i in extracted_list:
        if np.random.rand() > 0.5:
            noise_df.loc[i, column] = generate_random_value(upper_bound, new_upper_limit)
        else:
            noise_df.loc[i, column] = generate_random_value(new_lower_limit, lower_bound)

    return noise_df


def outlierDiscreteNew3Sigma(train_df,column,distribution,percentage,number,params):
  #extracted_list=sampleList(percentage,len(train_df[column]))
  extracted_list= sample_indices(distribution, train_df[column], percentage=None, number=None, params=None)
  noise_df= train_df.copy()
  mean = np.mean(noise_df[column])
  std_dev = np.std(noise_df[column])
  upper_bound=round(mean + 3 * std_dev,0)
  lower_bound=round(mean -3 * std_dev,0)
  new_upper_limit = round(mean + 4 * std_dev,0)
  new_lower_limit = round(mean - 4 * std_dev,0)
  if upper_bound==0:
     upper_bound=1
  if new_upper_limit==upper_bound:
     new_upper_limit=5*upper_bound

  if lower_bound==0:
     lower_bound=-1
  if new_lower_limit==lower_bound and lower_bound<0:
     new_lower_limit=5*lower_bound
  elif lower_bound>0:
     new_lower_limit=-5*lower_bound
  if upper_bound>new_upper_limit:
      tmp=new_upper_limit
      new_upper_limit=upper_bound
      upper_bound=tmp
  if new_lower_limit>lower_bound:
     tmp=lower_bound
     lower_bound=new_lower_limit
     new_lower_limit=tmp
  for i in extracted_list:
        if np.random.rand() > 0.5:
            noise_df.loc[i, column] = generate_random_value_discrete(upper_bound, new_upper_limit)
        else:
            noise_df.loc[i, column] = generate_random_value_discrete(new_lower_limit, lower_bound)

  return noise_df

def outlierCategoricalIntegerNew(train_df,column,distribution,percentage,number,params):
  #extracted_list=sampleList(percentage,len(train_df[column]))
  extracted_list= sample_indices(distribution, train_df[column], percentage=None, number=None, params=None)
  noise_df= train_df.copy()
  max_value = noise_df[column].max()
  min_value=noise_df[column].min()
  if min_value==0:
      min_value=-1
      new_minValue=-5
  elif min_value<0:
      new_minValue=5*min_value
  else:
      new_minValue=-5+min_value
  if max_value==0:
      max_value=1
      new_maxValue=5
  else:
      new_maxValue=2*max_value
  for i, value in enumerate(extracted_list):
     if np.random.rand() > 0.5:
      noise_df.loc[i, column] = np.random.randint(max_value, new_maxValue)
     else:
      noise_df.loc[i, column] = np.random.randint(new_minValue, min_value)

  return noise_df

def outlierCategoricalStringNew(train_df,column,distribution,percentage,number,params):
  #extracted_list=sampleList(percentage,len(train_df[column]))
  extracted_list= sample_indices(distribution, train_df[column], percentage=None, number=None, params=None)
  noise_df= train_df.copy()
  for i, value in enumerate(extracted_list):
    noise_df.loc[i, column] = "puck was here"
  return noise_df

# Funzione per identificare le date valide
def identify_valid_dates(df, date_column):
    valid_dates = []
    for date in df[date_column]:
        try:
            # Converte la data in stringa se necessario, prima di passarla a parser.parse
            if isinstance(date, (pd.Timestamp, date)):
                date = str(date)  # Converte in stringa
            parser.parse(date)
            valid_dates.append(True)
        except (ValueError, TypeError):
            valid_dates.append(False)

    df['is_valid_date'] = valid_dates
    return df

# Funzione per standardizzare solo le date riconosciute in un formato coerente
def standardize_dates(df, date_column):
    standardized_dates = []

    for date, is_valid in zip(df[date_column], df['is_valid_date']):
        if is_valid:
            try:
                # Converte la data con `dayfirst=True` e formatta come 'YYYY-MM-DD'
                standardized_date = parser.parse(date, dayfirst=True).date()  # Mantiene solo la parte della data
                standardized_dates.append(standardized_date)
            except (ValueError, TypeError):
                standardized_dates.append(date)  # Mantieni la data originale se c'è un errore
        else:
            standardized_dates.append(date)  # Mantieni la stringa originale per date non valide

    df[date_column] = standardized_dates
    return df

# Funzione per generare una data outlier basata su deviazione standard con un limite massimo
def random_outlier_date(mean_date, std_dev_days, min_date, max_date, factor=3, max_offset_days=36500):
    # Aumenta la variabilità generando un offset più casuale per ciascun outlier
    days_offset = random.choice([-1, 1]) * random.randint(factor, factor * 5) * random.uniform(0.5, 3) * std_dev_days

    # Limita il valore dell'offset per evitare l'errore OutOfBoundsTimedelta
    if abs(days_offset) > max_offset_days:
        days_offset = random.choice([-1, 1]) * random.randint(0, max_offset_days)  # Limita l'offset massimo

    # Calcola la nuova data outlier
    outlier_date = (mean_date + timedelta(days=days_offset)).date()

    return outlier_date

def random_upper_lower(series):
    """Apply random upper/lower casing to each string in the Series."""
    return series.apply(lambda text: ''.join(
        [char.upper() if random.random() > 0.5 else char.lower() for char in text]
    ))

def replace_punctuation(text):
    """Sostituisce ogni segno di punteggiatura (/, ., -) con uno degli altri due in maniera casuale, mantenendo lo stesso sostituto all'interno della stessa stringa."""
    replacements = {
        '/': random.choice(['-', '.']),
        '.': random.choice([',', '/']),
        '-': random.choice(['/', '.'])}
    return ''.join([replacements[char] if char in replacements else char for char in text])

def remove_or_replace(text):
    """Rimuove o sostituisce un carattere casuale, sostituendo lettere con lettere e numeri con numeri."""
    if len(text) < 2:
        return text
    pos = random.randint(0, len(text) - 1)
    char = text[pos]

    if char.isdigit():
        if random.random() > 0.5:
            return text[:pos] + text[pos + 1:]
        else:
            random_digit = str(random.randint(0, 9))
            return text[:pos] + random_digit + text[pos + 1:]
    elif char.isalpha():
        if random.random() > 0.5:
            return text[:pos] + text[pos + 1:]
        else:
            random_letter = random.choice(string.ascii_letters)
            return text[:pos] + random_letter + text[pos + 1:]
    return text

def abbreviate_text(text):
    """Abbrevia il testo mantenendo solo le prime lettere di ogni parola."""
    return ''.join([word[0] for word in text.split()])

def shuffle_words(text):
    """Cambia l'ordine delle parole in una stringa, garantendo un ordine diverso dall'originale."""
    words = text.split()
    while True:
        random.shuffle(words)
        if words != text.split():
            break
    return ' '.join(words)

# Funzione per creare JSON 
def create_json_for_feature(percentage, mode, feature_name=None, selection_criteria="all"):
    config = {
        "strategy": {
            "affected_features": [feature_name] if feature_name else [],
            "selection_criteria": selection_criteria,
            "percentage": percentage,
            "mode": mode,
            "perturbate_data": {
                "distribution": "random",
                "param": None,
                "value": None,
                "condition_logic": None
            }
        }
    }
    return config

def affectedSelected(affected_features,selection_criteria):
        affectedSelected=r'\b('+"|".join(map(re.escape,affected_features))+r')\b'
        found=re.findall(affectedSelected,selection_criteria)
        if found:
            modified=re.sub(affectedSelected,r'puck_\1',selection_criteria)
            return modified,1
        else:
            return selection_criteria,0



from pucktrick.utils import *

def noise(train_df, strategy, original_df=None):
 
    affected_features = strategy.get("affected_features")
    selection_criteria = strategy.get("selection_criteria")
    percentage = strategy.get("percentage")
    mode = strategy.get("mode")
    perturbate_config = strategy.get("perturbate_data", {})
    distr1 = perturbate_config.get("distribution")
    value = perturbate_config.get("value", [])
    condition_logic = perturbate_config.get("condition_logic", None)
    params=perturbate_config.get("param", {})
    isAffectedSelected=0
    if selection_criteria=='all':
            train1_df = train_df.copy()
            if original_df is None:
                original_df = train_df.copy()
            original1_df=original_df.copy()
    else:
        if mode=="extended":
            selection_criteria,isAffectedSelected=affectedSelected(affected_features,selection_criteria)
        else:
           isAffectedSelected=0 
        if isAffectedSelected==1:
            if original_df is not None:
                df_pref2=original_df[affected_features].add_prefix("puck_")
                original_df=pd.concat([original_df,df_pref2],axis=1)
                train_df=pd.concat([train_df,df_pref2],axis=1)
         
        if original_df is None:
            original1_df=  train_df.copy()
            original_df = train_df.query(selection_criteria).copy()         
        else:
            train1_df = train_df.query(selection_criteria).copy()
            original1_df=original_df.copy()
            original_df=original_df.query(selection_criteria).copy()
        train_df=train_df.query(selection_criteria).copy()
        noisy_df=train_df.copy()    
    
    if mode == "extended":
        train_df = original_df
    noisy_df = train_df.copy()
    if affected_features and not isinstance(affected_features, list):
        affected_features = [affected_features]

    if not value:
        filtered_df = noisy_df  
    else:
        row_filters = []
        for feature, value in zip(affected_features, value):
            if value is None:  
                row_filters.append(train_df[feature].notnull())
            else:
                row_filters.append(train_df[feature] == value)

        if condition_logic == "and":
            combined_filter = row_filters[0]
            for f in row_filters[1:]:
                combined_filter &= f
        elif condition_logic == "or":
            combined_filter = row_filters[0]
            for f in row_filters[1:]:
                combined_filter |= f
        elif condition_logic is None:
            combined_filter = row_filters[0]
            for f in row_filters[1:]:
                combined_filter &= f  
        else:
            raise ValueError("condition_logic must be either 'and', 'or', or None.")

        filtered_df = noisy_df[combined_filter]
    
    
    total_rows_to_change = int(len(filtered_df) * percentage)

    if mode == "extended" and original_df is not None:
        train_df = train1_df
        already_modified_mask = (train_df[affected_features[0]] != original_df[affected_features[0]])
        already_modified_count = already_modified_mask.sum()
        rows_to_change = total_rows_to_change - already_modified_count

        if rows_to_change > 0:
            modifiable_indices = filtered_df[~already_modified_mask.reindex(filtered_df.index, fill_value=False)]
            if len(modifiable_indices) == 0:
                print("No modifiable rows available in extended mode.")
                return 1,train_df
            number= rows_to_change
            percentage=None
            sampled_indices = sample_indices(distr1, modifiable_indices,percentage, number, params)
            sampled_indices = modifiable_indices.iloc[sampled_indices].index
        else:
            print("No additional rows to change in extended mode.")
            return 1,train_df

    else:  
        number= total_rows_to_change
        percentage=None
        sampled_indices = sample_indices(distr1, filtered_df,percentage, number, params)
        sampled_indices = filtered_df.iloc[sampled_indices].index

    if mode == "extended":
        noisy_df = train1_df

    
    for feature in affected_features:
        if pd.api.types.is_string_dtype(noisy_df[feature]):
            unique_values = train_df[feature].dropna().unique()
            for idx in sampled_indices:
                if len(unique_values)<=1:  
                   
                    noisy_df.loc[idx, feature] = ''.join(np.random.choice(list('abcdefghijklmnopqrstuvwxyz'), size=5))
                    
                else:
                    
                    while True:
                        new_value = np.random.choice(unique_values)
                        if new_value != noisy_df.loc[idx, feature]:
                            noisy_df.loc[idx, feature] = new_value
                            break

        elif pd.api.types.is_numeric_dtype(noisy_df[feature]):
            min_val = train_df[feature].min()
            max_val = train_df[feature].max()
            for idx in sampled_indices:
                if pd.api.types.is_integer_dtype(noisy_df[feature]):
                    new_value = random.randint(min_val, max_val)
                else:
                    new_value = random.uniform(min_val, max_val)
                noisy_df.loc[idx, feature] = new_value

        elif pd.api.types.is_datetime64_any_dtype(noisy_df[feature]):
            valid_dates = train_df[feature].dropna()
            if valid_dates.empty:
                continue
            min_date = valid_dates.min()
            max_date = valid_dates.max()
            for idx in sampled_indices:
                random_days = random.randint(0, (max_date - min_date).days)
                noisy_df.loc[idx, feature] = min_date + timedelta(days=random_days)

        elif pd.api.types.is_categorical_dtype(noisy_df[feature]):
            
            unique_values = train_df[feature].cat.categories
            for idx in sampled_indices:
                new_value = random.choice(unique_values)
                noisy_df.loc[idx, feature] = new_value

        else:
            raise ValueError(f"Unsupported column type for feature '{feature}'.")
    if isAffectedSelected==1:
        puck_cols=[c for c in noisy_df.columns if c.startswith("puck_")]
        noisy_df=noisy_df.drop(columns=puck_cols)
        original1_df=original1_df.drop(columns=puck_cols)
    
    original1_df.loc[noisy_df.index, noisy_df.columns]=noisy_df
    return 0,original1_df




import sys
from pathlib import Path

# aggiunge ../src al PYTHONPATH
ROOT = Path(__file__).resolve().parents[1]
SRC = ROOT / "src"
sys.path.insert(0, str(SRC))


import unittest
from pucktrick.noisy import *
from pucktrick.utils import * 
import pandas as pd
from pandas.testing import assert_frame_equal
class TestCategorical(unittest.TestCase):
    def test_saluta(self):
        percentage=0.5
        num_rows = 20
        selection_criteria = "f4=='apple'"
        selection_criteria='all'
        fake_df=create_fake_table(num_rows)
        column=['f3']
        print("orginal")
        print(fake_df)
        strategy =  {'affected_features': column, 'selection_criteria': selection_criteria, 'percentage': percentage, 'mode': 'new', 'perturbate_data': {'distribution': 'random', 'param': None, 'value': None, 'condition_logic': None}}
        error,noise_df=noise(fake_df,strategy)
        if error==1:
                print("error")
        print(" primo noisy")
        print(noise_df)
        
        percentage=0.7
        strategy =  {'affected_features': column, 'selection_criteria': selection_criteria, 'percentage': percentage, 'mode': 'extended', 'perturbate_data': {'distribution': 'random', 'param': None, 'value': None, 'condition_logic': None}}
        error,noise1_df=noise(noise_df,strategy,fake_df)
        print("error "+str(error))
        if error==0:
            noise_df=noise1_df
        print("secondo noisy")
        print(noise_df)
        if selection_criteria=='all':
            dif = fake_df[column] != noise_df[column]
            diff_number = dif.sum().sum()/len(column)
            print(diff_number)
            diff_p=diff_number/num_rows
            print(diff_p)
        else:
            selection_criteria,isAffected=affectedSelected(column,selection_criteria)
            if isAffected==1:
                df_pref2=fake_df[column].add_prefix("puck_")
                fake_df=pd.concat([fake_df,df_pref2],axis=1)
                noise_df=pd.concat([noise_df,df_pref2],axis=1)
            dif = fake_df.query(selection_criteria)[column] != noise_df.query(selection_criteria)[column]
            diff_number = dif.sum().sum()/len(column)
            diff_p=diff_number/len(fake_df.query(selection_criteria))
    
        # Verifica che il numero di modifiche sia corretto
        self.assertLessEqual(abs(diff_p-percentage),0.15)
        


if __name__ == "__main__":
    unittest.main()



from pucktrick.utils import *
def labels(train_df, strategy, original_df=None):

    target = strategy.get("affected_features")
    selection_criteria = strategy.get("selection_criteria")
    percentage = strategy.get("percentage")
    mode = strategy.get("mode")
    perturbate_config = strategy.get("perturbate_data", {})
    distr1 = perturbate_config.get("distribution")
    value = perturbate_config.get("value", [])
    condition_logic = perturbate_config.get("condition_logic", None)
    params= perturbate_config.get("param", {})
    number=None
    target_col = target[0] if isinstance(target, list) else target
    unique_values = train_df[target_col].dropna().unique()
    isAffectedSelected=0
    if selection_criteria=="all":
         
        if original_df is None:
            original_df = train_df.copy()
            original1_df=original_df
        else:
            original1_df=original_df.copy()
        
    else:
        if mode=="extended":
            selection_criteria,isAffectedSelected=affectedSelected(target,selection_criteria)
        else:
           isAffectedSelected=0 
        if isAffectedSelected==1:
            if original_df is not None:
                df_pref2=original_df[target].add_prefix("puck_")
                original_df=pd.concat([original_df,df_pref2],axis=1)
                train_df=pd.concat([train_df,df_pref2],axis=1)
         
        if original_df is None:
            original1_df=  train_df.copy()
            original_df = train_df.query(selection_criteria).copy()         
        else:
            original1_df=original_df.copy()
            original_df=original_df.query(selection_criteria).copy()
        train_df=train_df.query(selection_criteria).copy()
        

    #unique_values = train_df[target_col].dropna().unique()  
    if len(unique_values) == 2:
        var_type = "binary"
    elif len(unique_values) > 2:
        var_type = "categorical"
    else:
        raise ValueError(f"Unable to determine the type of target column '{target}'.")
    
   
    if var_type == "binary":
        #print("Target is binary")
        if mode == 'extended':
            noise_df= noiseBinaryExt(original_df, train_df, target_col,distr1, percentage,number,params)
        else:
            noise_df = noiseBinaryNew(train_df, target_col, distr1, percentage,params,number) 

    elif var_type == "categorical":
        if mode == 'extended':
            noise_df = noiseCategoricalIntExtendedExistingValues(original_df, train_df, target_col,distr1, percentage,number,params)

        else:
            noise_df = noiseCategoricalIntNewExistingValues(train_df, target_col, distr1, percentage,number,params) 
    for col in noise_df.select_dtypes(include=['int64']).columns:
        noise_df[col] = noise_df[col].astype('Int64')  
    if isAffectedSelected==1:
        puck_cols=[c for c in noise_df.columns if c.startswith("puck_")]
        noise_df=noise_df.drop(columns=puck_cols)
        original1_df=original1_df.drop(columns=puck_cols)
    
    original1_df.loc[noise_df.index, noise_df.columns]=noise_df
    return 0,original1_df



import sys
from pathlib import Path

# aggiunge ../src al PYTHONPATH
ROOT = Path(__file__).resolve().parents[1]
SRC = ROOT / "src"
sys.path.insert(0, str(SRC))


import unittest
from pucktrick.missing import *
from pucktrick.utils import *
import pandas as pd
import numpy as np
from pandas.testing import assert_frame_equal

class Test_outlies(unittest.TestCase):
  def test_MissingAll(self):
      percentage=0.5
      selection_criteria="f4=='apple'"
      #selection_criteria="all"
      num_rows = 20
      fake_df=create_fake_table(num_rows)
      column=['f3']
      strategy =  {'affected_features': column, 'selection_criteria': selection_criteria, 'percentage': percentage, 'mode': 'new', 'perturbate_data': {'distribution': 'random', 'param': None, 'value': None, 'condition_logic': None}}
      print(fake_df)
      error,noise_df=missing(fake_df,strategy)
      if error==1:
                print("error")
      print(" primo noisy")
      print(noise_df)
      percentage=0.7
      strategy =  {'affected_features': column, 'selection_criteria': selection_criteria, 'percentage': percentage, 'mode': 'extended', 'perturbate_data': {'distribution': 'random', 'param': None, 'value': None, 'condition_logic': None}}
      error, noise1_df=missing(noise_df,strategy,fake_df)
      print("error "+str(error))
      if error==0:
            noise_df=noise1_df
      print("secondo noisy")
      print(noise_df)
      if selection_criteria == "all" or selection_criteria is None:
            dif = fake_df[column] != noise_df[column]
            diff_number = dif.sum().sum()/len(column)
            diff_p=round(diff_number/num_rows,1)
      else:
            selection_criteria,isAffected=affectedSelected(column,selection_criteria)
            if isAffected==1:
                df_pref2=fake_df[column].add_prefix("puck_")
                fake_df=pd.concat([fake_df,df_pref2],axis=1)
                noise_df=pd.concat([noise_df,df_pref2],axis=1)
            dif = fake_df.query(selection_criteria)[column] != noise_df.query(selection_criteria)[column]
            diff_number = dif.sum().sum()/len(column)
            diff_p=round(diff_number/len(fake_df.query(selection_criteria)),1)
      print("diff_p", diff_p)
      print("percentage", percentage)
      self.assertLessEqual(abs(diff_p-percentage),0.15)
  
if __name__ == "__main__":
    unittest.main()
